USE TEMPDB
GO

CREATE TABLE dbo.Hierarchy (
			id int not null,
			parentid int,
			nodename varchar(50),
			nodetype varchar(20) )
GO

INSERT INTO dbo.Hierarchy ( id, parentid, nodename, nodetype )
values (1,	NULL,	'Seattle Grace Hospital',	'Hospital'),
(2,	1,	'Cardiology',	'Department'),
(3,	1,	'General Surgery',	'Department'),
(4,	2,	'Preston Burke',	'Consultant'),
(5,	2,	'Cristina Yang',	'Resident'),
(6,	2,	'Erica Hahn',	'Consultant'),
(7,	3,	'Richard Webber',	'Consultant'),
(8,	3,	'Miranda Bailey',	'Consultant'),
(9,	3,	'Meredith Grey',	'Resident')
GO

-- SQL 2005 introduced CTEs which are great for recursion
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'[dbo].[cte_getchildren]') AND type in (N'P', N'PC'))
DROP PROCEDURE [dbo].[cte_getchildren]
GO

CREATE PROCEDURE dbo.cte_getchildren (@currentid int) as

SET NOCOUNT ON ; 

WITH CTEhierarchy (id, nodename, [Level])
AS
(
-- Anchor member definition
    SELECT h.id, h.nodename, 0 AS [Level]
    FROM dbo.Hierarchy AS h
    WHERE h.id = @currentid
    UNION ALL
-- Recursive member definition
    SELECT c.id, c.nodename, [Level]+1
    FROM dbo.Hierarchy AS c
    INNER JOIN CTEhierarchy AS p ON c.parentid = p.id
)
-- Statement that executes the CTE
SELECT id, nodename, [Level], REPLICATE('  |  ',[Level]) + nodename AS [Hierachy]
FROM CTEhierarchy
GO

-- This is what it does!
EXEC dbo.cte_getchildren 2 -- Cardiology
GO

-- The whole lot!
EXEC dbo.cte_getchildren 1 -- Seattle Grace Hospital
GO

-- Create Materialised Path from existing Ids
WITH CTEhierarchy (id, [Path], [Level] )
AS
(
-- Anchor member definition
    SELECT h.id, CAST('/' + CAST(h.id as varchar(10)) + '/' AS varchar(max)) AS [Path], 0 AS [Level]
	FROM dbo.Hierarchy AS h
    WHERE h.id = 1 -- Root Node
    UNION ALL
-- Recursive member definition
    SELECT c.id, p.Path + CAST(c.id as varchar(10)) + '/' AS [Path], [Level] + 1 AS [Level]
    FROM dbo.Hierarchy AS c
    INNER JOIN CTEhierarchy AS p ON c.parentid = p.id
)
-- Statement that executes the CTE
SELECT c.id, c.[Path], h.parentid, h.nodename, h.nodetype, c.[Level], REPLICATE('  |  ',c.[Level]) + h.nodename AS [Hierachy]
FROM dbo.Hierarchy h
JOIN CTEhierarchy c ON c.id = h.id
GO

-- Create OrdPath
WITH CTEhierarchy (id, [OrdPath], [Level])
AS
(
-- Anchor member definition
    SELECT h.id, CAST('/' AS varchar(max)) AS [OrdPath], 0 AS [Level]
	FROM dbo.Hierarchy AS h
    WHERE h.id = 1 -- Root Node
    UNION ALL
-- Recursive member definition
    SELECT c.id, p.OrdPath + CAST(ROW_NUMBER() OVER ( PARTITION BY c.parentid ORDER BY c.nodename, c.id ) AS varchar(10)) + '/' AS [OrdPath], [Level]+1 AS [Level]
    FROM dbo.Hierarchy AS c
    INNER JOIN CTEhierarchy AS p ON c.parentid = p.id
)
-- Statement that executes the CTE
SELECT c.id, c.[OrdPath], h.parentid, h.nodename, h.nodetype, c.[Level], REPLICATE('  |  ',c.[Level]) + h.nodename AS [Hierachy]
FROM dbo.Hierarchy h
JOIN CTEhierarchy c ON c.id = h.id
ORDER BY c.[OrdPath] 
GO

ALTER TABLE dbo.Hierarchy ADD hid HierarchyId
GO

-- We can use this as an update to populate HierarchyId
WITH CTEhierarchy (id, [OrdPath])
AS
(
-- Anchor member definition
    SELECT h.id, CAST('/' AS varchar(max)) AS [OrdPath]
	FROM dbo.Hierarchy AS h
    WHERE h.id = 1 -- Root Node
    UNION ALL
-- Recursive member definition
    SELECT c.id, p.OrdPath + CAST(ROW_NUMBER() OVER ( PARTITION BY c.parentid ORDER BY c.nodename, c.id ) AS varchar(10)) + '/' AS [OrdPath]
    FROM dbo.Hierarchy AS c
    INNER JOIN CTEhierarchy AS p ON c.parentid = p.id
)
-- Statement that executes the CTE
-- Set the HierarchyId value
UPDATE h
SET hid = c.[OrdPath] -- Implicit conversion could use hierarchyId::Parse(c.[OrdPath])
FROM dbo.Hierarchy h
JOIN CTEhierarchy c ON c.id = h.id
GO

ALTER TABLE dbo.Hierarchy ALTER COLUMN hid HierarchyId NOT NULL 
GO

ALTER TABLE dbo.Hierarchy ADD
CONSTRAINT PK_Hierarchy  PRIMARY KEY CLUSTERED (Hid)
GO

-- Using a hierachyId
-- Parent/Child Queries Important so Clustered PK
CREATE TABLE HospitalHierarchy ( hid HierarchyId NOT NULL CONSTRAINT PK_HospitalHierarchy PRIMARY KEY CLUSTERED,
						[level] AS hid.GetLevel(),
						name varchar(30) not null,
						[type] char(10) not null
						)
GO

-- Create a breath-first index
CREATE UNIQUE INDEX HospitalHierarchy_Breadth_Idx ON HospitalHierarchy(Level,hid) ;
GO


INSERT HospitalHierarchy ( hid,
			name,
			type)
VALUES ( hierarchyid::GetRoot(),
		'Seatle Grace Hospital',
		'Hospital' )
GO

-- Make sure in Grid Mode to return what has been inserted
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO
		
-- To get the root node from any hierarchy use GetRoot() and test against column value
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
WHERE hid = hierarchyid::GetRoot()
GO

-- Add the next levels
-- GetRoot() gets the Root of the Hierarchy Tree
-- GetDescendant() returns a child node between two children
-- Rules for GetDescendant()
-- 1. If parent is NULL, returns NULL.
-- 2. If parent is not NULL, and both child1 and child2 are NULL, returns a child of parent.
-- 3. If parent and child1 are not NULL, and child2 is NULL, returns a child of parent greater than child1.
-- 4. If parent and child2 are not NULL and child1 is NULL, returns a child of parent less than child2.
-- 5. If parent, child1, and child2 are all not NULL, returns a child of parent greater than child1 and less than child2.
-- 6. If child1 or child2 is not NULL but is not a child of parent, an exception is raised.
-- 7. If child1 >= child2, an exception is raised.

-- Case 2
DECLARE @RootNode hierarchyid 
SELECT @RootNode = hierarchyid::GetRoot()
SELECT @RootNode

INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( @RootNode.GetDescendant(NULL, NULL), 
		'Cardiology',
		'Department' )
GO

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

-- Case 3
-- Add a child after 'Cardiology'
DECLARE @RootNode hierarchyid 
DECLARE @Child1 hierarchyid 
SELECT @RootNode = hierarchyid::GetRoot()
SELECT @Child1 = hid
FROM HospitalHierarchy
WHERE name = 'Cardiology'

INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( @RootNode.GetDescendant(@Child1, NULL), 
		'General Surgery',
		'Department' )

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

-- Case 5
-- Add a child between 'Cardiology' and 'General Surgery',
DECLARE @RootNode hierarchyid 
DECLARE @Child1 hierarchyid 
DECLARE @Child2 hierarchyid 
SELECT @RootNode = hierarchyid::GetRoot()
SELECT @Child1 = hid
FROM HospitalHierarchy
WHERE name = 'Cardiology'
SELECT @Child2 = hid
FROM HospitalHierarchy
WHERE name = 'General Surgery'

INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( @RootNode.GetDescendant(@Child1, @Child2), 
		'Genito-Urinary Medicine',
		'Department' )
GO

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

-- Case 4
-- Add a child before 'Cardiology'
DECLARE @RootNode hierarchyid 
DECLARE @Child2 hierarchyid 
SELECT @RootNode = hierarchyid::GetRoot()
SELECT @Child2 = hid
FROM HospitalHierarchy
WHERE name = 'Cardiology'

INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( @RootNode.GetDescendant(NULL, @Child2), 
		'Obstetrics and Gynaecology',
		'Department' )
GO

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

-- Note Path for Obstetrics and Gynaecology is '/0/'
-- If try to insert before that then you get /-1/
DECLARE @RootNode hierarchyid 
DECLARE @Child2 hierarchyid 
SELECT @RootNode = hierarchyid::GetRoot()
SELECT @Child2 = hid
FROM HospitalHierarchy
WHERE name = 'Obstetrics and Gynaecology'

SELECT @RootNode.GetDescendant(NULL, @Child2).ToString()
GO

-- If we wanted to add a node anywhere we have a problem! (PK Violation)
DECLARE @RootNode hierarchyid 
SELECT @RootNode = hierarchyid::GetRoot()
SELECT @RootNode.GetDescendant(NULL, NULL).ToString() 
GO

-- See if we can add a node after 'General Surgery'
DECLARE @RootNode hierarchyid 
DECLARE @Child1 hierarchyid 
SELECT @RootNode = hierarchyid::GetRoot()
SELECT @Child1 = hid
FROM HospitalHierarchy
WHERE name = 'General Surgery'
SELECT @RootNode.GetDescendant(@Child1, NULL).ToString() 
GO

-- Add a node after 'General Surgery'
DECLARE @RootNode hierarchyid 
DECLARE @Child1 hierarchyid 
SELECT @RootNode = hierarchyid::GetRoot()
SELECT @Child1 = hid
FROM HospitalHierarchy
WHERE name = 'General Surgery'

INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( @RootNode.GetDescendant(@Child1, NULL), 
		'Neurology',
		'Department' )

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

-- Specify specific node using Parse
DECLARE @hid hierarchyId
SET @hid = hierarchyId::Parse('/4/') -- Could have done CAST ( '/4/' AS hierarchyId )

INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( @hid ,
		'Orthopaedics',
		'Department' )

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

DECLARE @RootNode hierarchyid 
DECLARE @Child1 hierarchyid 
SELECT @RootNode = hierarchyid::GetRoot()

INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( @RootNode.GetDescendant((SELECT MAX(hid) FROM HospitalHierarchy), NULL), 
		'Reconstructive Surgery',
		'Department' )
GO

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

-- Add a node under 'General Surgery'
INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( (SELECT hid FROM HospitalHierarchy WHERE name = 'General Surgery').GetDescendant(NULL, NULL), 
				'Richard Webber',
				'Consultant' )
GO

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

-- Using IsDescendant
SELECT p.hid.ToString() AS [Parent], c.hid.ToString() AS [Child]
FROM HospitalHierarchy p
JOIN HospitalHierarchy c ON p.hid.IsDescendant(c.hid) = 1
GO
-- Note: You are decendent  of yourself!
--				this method may become IsDecendentOf()

INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( (SELECT hid FROM HospitalHierarchy WHERE name = 'General Surgery').GetDescendant(
							( SELECT MAX(c.hid) FROM HospitalHierarchy p JOIN HospitalHierarchy c ON p.hid.IsDescendant(c.hid) = 1 AND  p.name = 'General Surgery')
							, NULL), 
				'Miranda Bailey',
				'Consultant' )
GO

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( (SELECT hid FROM HospitalHierarchy WHERE name = 'General Surgery').GetDescendant(
							( SELECT MAX(c.hid) FROM HospitalHierarchy p JOIN HospitalHierarchy c ON p.hid.IsDescendant(c.hid) = 1 AND  p.name = 'General Surgery')
							, NULL), 
				'Meredith Grey ',
				'Resident' )
GO

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

-- Add a node under 'Neurology'
INSERT HospitalHierarchy ( hid,
			name,
			type )
VALUES ( (SELECT hid FROM HospitalHierarchy WHERE name = 'Neurology').GetDescendant(NULL, NULL), 
				'Derek Shepherd',
				'Consultant' )
GO

-- Show hierarchy
SELECT hid.ToString() as node,
			[level],
			name,
			[type]
FROM HospitalHierarchy
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
SELECT  (SELECT hid FROM HospitalHierarchy WHERE name = 'Cardiology').GetDescendant( NULL, NULL), 
				'Preston Burke',
				'Consultant' 
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
SELECT  (SELECT hid FROM HospitalHierarchy WHERE name = 'Cardiology').GetDescendant( (SELECT MAX(c.hid) FROM HospitalHierarchy p JOIN HospitalHierarchy c ON p.hid.IsDescendant(c.hid) = 1 AND  p.name = 'Cardiology'), NULL), 
				'Cristina Yang',
				'Resident' 
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
SELECT  (SELECT hid FROM HospitalHierarchy WHERE name = 'Cardiology').GetDescendant( (SELECT MAX(c.hid) FROM HospitalHierarchy p JOIN HospitalHierarchy c ON p.hid.IsDescendant(c.hid) = 1 AND  p.name = 'Cardiology'), NULL), 
				'Erica Hahn ',
				'Consultant' 
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
SELECT  (SELECT hid FROM HospitalHierarchy WHERE name = 'Obstetrics and Gynaecology').GetDescendant( NULL, NULL), 
				'Addison Montgomery',
				'Consultant' 
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
SELECT  (SELECT hid FROM HospitalHierarchy WHERE name = 'Obstetrics and Gynaecology').GetDescendant( (SELECT MAX(c.hid) FROM HospitalHierarchy p JOIN HospitalHierarchy c ON p.hid.IsDescendant(c.hid) = 1 AND  p.name = 'Obstetrics and Gynaecology'), NULL), 
				'Izzie Stevens',
				'Resident' 
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
SELECT  (SELECT hid FROM HospitalHierarchy WHERE name = 'Reconstructive Surgery').GetDescendant( NULL, NULL), 
				'Mark Sloan',
				'Consultant' 
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
SELECT  (SELECT hid FROM HospitalHierarchy WHERE name = 'Reconstructive Surgery').GetDescendant( (SELECT MAX(c.hid) FROM HospitalHierarchy p JOIN HospitalHierarchy c ON p.hid.IsDescendant(c.hid) = 1 AND  p.name = 'Reconstructive Surgery'), NULL), 
				'Alex Karev',
				'Resident' 
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
SELECT  (SELECT hid FROM HospitalHierarchy WHERE name = 'Genito-Urinary Medicine').GetDescendant( NULL, NULL), 
				'George O''Malley',
				'Intern' 
GO

INSERT HospitalHierarchy ( hid,
			name,
			type )
SELECT  (SELECT hid FROM HospitalHierarchy WHERE name = 'Orthopaedics').GetDescendant( NULL, NULL), 
				'Callie Torres',
				'Resident' 
GO

-- Show hierarchy
SELECT hid.ToString() as node,
		[Level],
			name,
			[Type]
FROM HospitalHierarchy
ORDER BY hid
GO

-- Get all children of Cardiology
DECLARE @FirstNode hierarchyid 
SELECT @FirstNode = hid
FROM HospitalHierarchy
WHERE name = 'Cardiology'

SELECT hid.ToString() as node,
		[Level],
			name,
			[Type]
FROM HospitalHierarchy
WHERE @FirstNode.IsDescendant(hid) = 1
ORDER BY hid
GO

-- Get the Parent (ancestor) of a node
DECLARE @FirstNode hierarchyid 
SELECT @FirstNode = hid
FROM HospitalHierarchy
WHERE name = 'Cristina Yang'

SELECT hid.ToString() as node,
		[Level],
			name,
			[Type]
FROM HospitalHierarchy
WHERE @FirstNode.GetAncestor (1) = hid
ORDER BY hid
GO

-- Get the Grand Parent of the node
DECLARE @FirstNode hierarchyid 
SELECT @FirstNode = hid
FROM HospitalHierarchy
WHERE name = 'Cristina Yang'

SELECT hid.ToString() as node,
		[Level],
			name,
			[Type]
FROM HospitalHierarchy
WHERE @FirstNode.GetAncestor (2) = hid
ORDER BY hid
GO

-- Using the breadth index
SELECT hid.ToString() as node,
		[Level]
FROM HospitalHierarchy
WHERE level = 2
ORDER BY hid
GO

-- Move consultants under Cardiology 
DECLARE @OldParent hierarchyid , @NewParent  hierarchyid
SELECT @OldParent = hid 
FROM HospitalHierarchy
WHERE name = 'Cardiology'

SELECT hid.ToString() as node,
		[Level],
			name,
			[Type]
FROM HospitalHierarchy
WHERE hid.GetAncestor (1) = @OldParent
GO
-- Move consultants under Cardiology to General Surgery
DECLARE @OldParent hierarchyid , @NewParent  hierarchyid
SELECT @OldParent = hid 
FROM HospitalHierarchy
WHERE name = 'Cardiology'
SELECT @NewParent = hid 
FROM HospitalHierarchy
WHERE name = 'General Surgery'

UPDATE HospitalHierarchy
SET hid = hid.Reparent(@OldParent, @NewParent) 
WHERE hid.GetAncestor (1) = @OldParent

-- THIS FAILS WITH A PK Violation Why?

DECLARE @OldParent hierarchyid , @NewParent  hierarchyid
SELECT @OldParent = hid 
FROM HospitalHierarchy
WHERE name = 'Cardiology'
SELECT @NewParent = hid 
FROM HospitalHierarchy
WHERE name = 'General Surgery'

SELECT hid.ToString(), hid.Reparent(@OldParent, @NewParent).ToString()
FROM HospitalHierarchy
WHERE hid.GetAncestor (1) = @OldParent

-- Reparent is giving a node that already exists
SELECT hid.ToString(),*
FROM HospitalHierarchy
WHERE hid = hierarchyid::Parse('/2/1/')


DECLARE @OldParent hierarchyid , @NewParent  hierarchyid, @increment int, @OldParentNode varchar(20)
SELECT @OldParent = hid , @OldParentNode = hid.ToString()
FROM HospitalHierarchy
WHERE name = 'Cardiology'
SELECT @NewParent = hid
FROM HospitalHierarchy
WHERE name = 'General Surgery'

SELECT @increment = CAST(SUBSTRING(node, LEN(NODE)-CHARINDEX('/',REVERSE(LEFT(Node,LEN(node)-1)))+1, CHARINDEX('/',REVERSE(LEFT(Node,LEN(node)-1)))-1) AS int)
FROM ( SELECT MAX(c.hid).ToString() AS node
FROM HospitalHierarchy p JOIN HospitalHierarchy c ON p.hid.IsDescendant(c.hid) = 1 AND  p.hid = @NewParent ) d

SELECT hid.ToString(), @OldParentNode + CAST( CAST(SUBSTRING(node, LEN(NODE)-CHARINDEX('/',REVERSE(LEFT(Node,LEN(node)-1)))+1, CHARINDEX('/',REVERSE(LEFT(Node,LEN(node)-1)))-1) AS int) + @increment  AS varchar(20) ) + '/'
FROM ( SELECT hid, hid.ToString() as node
FROM HospitalHierarchy
WHERE hid.GetAncestor (1) = @OldParent ) o

-- Now join it all together in an update
DECLARE @OldParent hierarchyid , @NewParent  hierarchyid, @increment int, @OldParentNode varchar(20), @NewParentNode varchar(20)
SELECT @OldParent = hid , @OldParentNode = hid.ToString()
FROM HospitalHierarchy
WHERE name = 'Cardiology'
SELECT @NewParent = hid, @NewParentNode = hid.ToString()
FROM HospitalHierarchy
WHERE name = 'General Surgery'

SELECT @increment = CAST(SUBSTRING(node.ToString(),LEN(@NewParentNode)+1, LEN(node.ToString())-1-LEN(@NewParentNode)) AS int)
FROM ( SELECT MAX(c.hid) AS node
FROM HospitalHierarchy p JOIN HospitalHierarchy c ON p.hid.IsDescendant(c.hid) = 1 AND  p.hid = @NewParent ) d

UPDATE h
SET hid = hierarchyid::Parse(@OldParentNode + CAST( CAST(SUBSTRING(h.hid.ToString(),LEN(@OldParentNode)+1, LEN(h.hid.ToString())-1-LEN(@OldParentNode)) AS int) + @increment  AS varchar(20) ) + '/').Reparent(@OldParent, @NewParent)
FROM HospitalHierarchy h
WHERE h.hid.GetAncestor (1) = @OldParent

-- Show hierarchy
SELECT hid.ToString() as node,
		[Level],
			name,
			[Type]
FROM HospitalHierarchy
ORDER BY hid

-- Clean Up
DROP TABLE HospitalHierarchy